
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "mcucmd.h"
#include "mcuserial.h"
#include "log.h"

#include <QtCore>




/********************************************************************/
/*!
    \class IMcuCmd
    \brief Base for MCU serial communication command

    This class is an approach to generically deal with the
    variety of different command and answer types.
*/


/*!
    \fn bool IMcuCmd::flush() const
    Return \c true if pending bytes for read from serial port
    should be flushed before sending command.

    The default implementation returns \c true.
*/


/*!
    \fn QByteArray IMcuCmd::command() const
    Get raw command string for submission to the MCU.
*/


/*!
    \fn int IMcuCmd::receive(const QByteArray &aAnswerFragment)

    McuSerial uses this method to hand over raw characters
    received from the MCU in \a aAnswerFragment.

    The return value indicates how much more answer
    characters are expected. It can change between
    subsequent calls of receive(), since some commands
    return a status character and, if status is not ok,
    additional descriptive characters. A return value of
    \c 0 indicates that no more characters are expected.
*/


/*!
    \fn QString IMcuCmd::name() const
    Return short human-readable name of command for debugging.
*/




/********************************************************************/
/*!
    \class McuAnsHelper
    \brief Convenience class for parsing answer bytearrays
*/

class McuAnsHelper
    {
public:
    McuAnsHelper(const QByteArray &aAns)
        : m_aAns(aAns) {};
    int byte(int nIx) const;
    int word(int nIx) const;
private:
    QByteArray m_aAns;
    };


int McuAnsHelper::byte(int nIx) const
    {
    if ((nIx < 0) || (nIx >= m_aAns.size()))
        return 0;
    return (int)(unsigned char)m_aAns[nIx];
    }


int McuAnsHelper::word(int nIx) const
    {
    if ((nIx < 0) || (nIx+1 >= m_aAns.size()))
        return 0;
    return (int)(unsigned char)m_aAns[nIx] +
        (((int)(unsigned char)m_aAns[nIx+1]) << 8);
    }




/********************************************************************/
/*!
    \class McuCmdVoid
    \brief Generic base for commands with no answer
*/

class McuCmdVoid : public IMcuCmd
    {
public:
    McuCmdVoid(const QByteArray &aCmd,
        const QString &qsName, bool bPreFlush = true)
        : m_aCmd(aCmd), m_qsName(qsName), m_bFlush(bPreFlush) {};

private:
    bool flush() const { return m_bFlush; };
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    QString m_qsName;
    bool m_bFlush;
    };


int McuCmdVoid::receive(const QByteArray &aAnswerFragment)
    {
    Q_UNUSED(aAnswerFragment)
    return 0;
    }




/********************************************************************/
/*!
    \class McuCmdBool
    \brief Generic base for commands with boolean answer and error string
*/

class McuCmdBool : public IMcuCmd
    {
public:
    McuCmdBool(const QByteArray &aCmd,
        const QString &qsName, char cSuccess = '0', bool bPreFlush = true)
        : m_aCmd(aCmd), m_aAns(), m_qsName(qsName),
          m_cSuccess(cSuccess), m_bFlush(bPreFlush) {};
    QByteArray answer() const { return m_aAns; };
    bool success() const;
    QString errorReason() const;

private:
    bool flush() const { return m_bFlush; };
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    QByteArray m_aAns;
    QString m_qsName;
    char m_cSuccess;
    bool m_bFlush;
    };


bool McuCmdBool::success() const
    {
    return (m_aAns.length() > 0) && (m_aAns[0] == m_cSuccess);
    }


QString McuCmdBool::errorReason() const
    {
    if ((m_aAns.length() < 1) || (m_aAns[0] == m_cSuccess))
        return QString();
    return QString::fromLatin1(m_aAns.mid(1));
    }


int McuCmdBool::receive(const QByteArray &aAnswerFragment)
    {
    // wait for at least 1 answer character
    // after that, keep receiving if first char is not m_cSuccess and if not delimited by #
    // until no further data is available
    m_aAns += aAnswerFragment;
    if (m_aAns.length() == 0)
        return 1;
    else if (((m_aAns.length() == 1) && (m_aAns[0] == m_cSuccess)) || (m_aAns.endsWith("#")))
        return 0;
    else
        return 1000;
    }




/********************************************************************/
/*!
    \class McuCmdString
    \brief Generic base for commands with string answer (delimited by #)
*/

class McuCmdString : public IMcuCmd
    {
public:
    McuCmdString(const QByteArray &aCmd,
        const QString &qsName, bool bPreFlush = true)
        : m_aCmd(aCmd), m_aAns(), m_qsName(qsName),
          m_bFlush(bPreFlush) {};
    QByteArray answer() const { return m_aAns; };
    bool success() const;
    QString errorReason() const;

private:
    bool flush() const { return m_bFlush; };
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    QByteArray m_aAns;
    QString m_qsName;
    bool m_bFlush;
    };


int McuCmdString::receive(const QByteArray &aAnswerFragment)
    {
    // wait for at least 1 answer character
    // after that, keep receiving if string not delimited by #
    // until no further data is available
    m_aAns += aAnswerFragment;
    if (m_aAns.length() == 0)
        return 1;
    else if (m_aAns.endsWith("#"))
        return 0;
    else
        return 1000;
    }




/********************************************************************/
/*!
    \class McuCmdFixed
    \brief Generic base for commands with fixed lenght answer
*/

class McuCmdFixed : public IMcuCmd
    {
public:
    McuCmdFixed(const QByteArray &aCmd, int nAnsLen,
        const QString &qsName, bool bPreFlush = true)
        : m_aCmd(aCmd), m_nAnsLen(nAnsLen), m_aAns(),
          m_qsName(qsName), m_bFlush(bPreFlush) {};
    QByteArray answer() const { return m_aAns; };

private:
    bool flush() const { return m_bFlush; };
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    int m_nAnsLen;
    QByteArray m_aAns;
    QString m_qsName;
    bool m_bFlush;
    };


int McuCmdFixed::receive(const QByteArray &aAnswerFragment)
    {
    // wait for at least m_nAnsLen answer characters
    m_aAns += aAnswerFragment;
    return m_nAnsLen - m_aAns.length();
    }




/********************************************************************/
/*!
    \class McuCmdQueryUnlimited
    \brief Generic base for querying undelimited strings
*/

class McuCmdQueryUnlimited : public IMcuCmd
    {
public:
    McuCmdQueryUnlimited(const QByteArray &aCmd, const QString &qsName)
        : m_aCmd(aCmd), m_aAns(), m_qsName(qsName) {};
    QByteArray answer() const { return m_aAns; };

private:
    QByteArray command() const { return m_aCmd; };
    int receive(const QByteArray &aAnswerFragment);
    QString name() const { return m_qsName; };

protected:
    QByteArray m_aCmd;
    QByteArray m_aAns;
    QString m_qsName;
    };


int McuCmdQueryUnlimited::receive(const QByteArray &aAnswerFragment)
    {
    // wait for at least 1 answer character
    // after that, keep receiving until no further characters are available
    m_aAns += aAnswerFragment;
    if (m_aAns.isEmpty())
        return 1;
    else if (aAnswerFragment.isEmpty())
        return 0;
    else
        return 1000;
    }




/********************************************************************/
/*!
    \class McuRawCommand
*/

bool McuRawCommand::mcuSend(McuSerial &rPort)
    {
    McuCmdQueryUnlimited oCmd(m_aCommand,
        QString::fromUtf8("RawCommand"));
    if (!rPort.send(&oCmd))
        return false;
    m_aAnswer = oCmd.answer();
    return true;
    }


/********************************************************************/
/*!
    \class McuDataFirmware
*/

bool McuDataFirmware::mcuGet(McuSerial &rPort)
    {
    m_qsFirmware = QString();
    McuCmdQueryUnlimited oCmd(
        QString::fromUtf8("#:GV#").toLatin1().constData(),
        QString::fromUtf8("GetFirmware"));
    if (!rPort.send(&oCmd))
        return false;
    m_qsFirmware = QString::fromLatin1(oCmd.answer().constData());
    return true;
    }


/********************************************************************/
/*!
    \class McuDataState
*/

bool McuDataState::mcuGet(McuSerial &rPort)
    {
    m_aState.clear();
    McuCmdFixed oCmd1(
        QString::fromUtf8("<#:VX#").toLatin1().constData(), 12,
        QString::fromUtf8("GetState1"));
    McuCmdFixed oCmd2(
        QString::fromUtf8(">#:VX#").toLatin1().constData(), 6,
        QString::fromUtf8("GetState1"));
    if ((!rPort.send(&oCmd1)) || (!rPort.send(&oCmd2)))
        return false;
    m_aState = oCmd1.answer() + oCmd2.answer();
    return true;
    }

QString McuDataState::backlash() const
    {
    McuAnsHelper oAns(m_aState);
    int nRaBacklashMethod = oAns.byte(0);
    int nRaBacklashVal = oAns.word(1);
    int nDecBacklashMethod = oAns.byte(3);
    int nDecBacklashVal = oAns.word(4);
    return QString::fromUtf8("RA %1, %2; Dec %3, %4")
        .arg(QString::fromUtf8(
            (nRaBacklashMethod == 0xf0)? "-1 (finish Left)":
            (nRaBacklashMethod == 0x00)?  "0 (none)":
            (nRaBacklashMethod == 0x0f)?  "1 (finish Right)": "?"))
        .arg(nRaBacklashVal)
        .arg(QString::fromUtf8(
            (nDecBacklashMethod == 0xf0)? "-1 (finish Down)":
            (nDecBacklashMethod == 0x00)?  "0 (none)":
            (nDecBacklashMethod == 0x0f)?  "1 (finish Up)": "?"))
        .arg(nDecBacklashVal);
    }

QString McuDataState::guideRate() const
    {
    // guiding rate (and paddle 2x)
    switch (McuAnsHelper(m_aState).byte(6))
        {
        case 0x01: return QString::fromUtf8("1.33x");
        case 0x02: return QString::fromUtf8("1.66x");
        case 0x03: return QString::fromUtf8("2.0x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::paddle8Rate() const
    {
    // center rate (paddle 8x)
    switch (McuAnsHelper(m_aState).byte(7))
        {
        case 0x07: return QString::fromUtf8("4x");
        case 0x08: return QString::fromUtf8("8x");
        case 0x09: return QString::fromUtf8("16x");
        case 0x0a: return QString::fromUtf8("20x");
        case 0x0b: return QString::fromUtf8("27x");
        case 0x0c: return QString::fromUtf8("40x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::paddle16Rate() const
    {
    // slew rate (paddle 16x)
    switch (McuAnsHelper(m_aState).byte(8))
        {
        case 0x07: return QString::fromUtf8("4x");
        case 0x08: return QString::fromUtf8("8x");
        case 0x09: return QString::fromUtf8("16x");
        case 0x0a: return QString::fromUtf8("20x");
        case 0x0b: return QString::fromUtf8("27x");
        case 0x0c: return QString::fromUtf8("40x");
        default: return QString::fromUtf8("???");
        }
    }

QString McuDataState::gotoRate() const
    {
    // goto rate
    switch (McuAnsHelper(m_aState).byte(9))
        {
        case 0x07: return QString::fromUtf8("4x");
        case 0x08: return QString::fromUtf8("8x");
        case 0x09: return QString::fromUtf8("16x");
        case 0x0a: return QString::fromUtf8("20x");
        case 0x0b: return QString::fromUtf8("27x");
        case 0x0c: return QString::fromUtf8("40x");
        default: return QString::fromUtf8("???");
        }
    }

int McuDataState::rotation() const
    {
    McuAnsHelper oAns(m_aState);
    return ((oAns.byte(10) & 0x0f) == 0x00)? 0: 1;
    }

int McuDataState::orientation() const
    {
    McuAnsHelper oAns(m_aState);
    return ((oAns.byte(10) & 0xf0) == 0x00)? 0: 1;
    }

QString McuDataState::rotOrient() const
    {
    return QString::fromUtf8("RA %1, Orientation %2")
        .arg(QString::fromUtf8((rotation() == 0)? "Original": "Reverse"))
        .arg(QString::fromUtf8((orientation() == 0)? "Left": "Right"));
    }

int McuDataState::trackingRate() const
    {
    switch (McuAnsHelper(m_aState).byte(11))
        {
        case  0: return 0; // sidereal
        case  1: return 1; // solar
        case  2: return 2; // lunar
        case -1: return 3; //off
        default: return 3;
        }
    }

int McuDataState::gearboxReduction() const
    {
    return McuAnsHelper(m_aState).word(12);
    }

int McuDataState::wormRatio() const
    {
    return McuAnsHelper(m_aState).word(14);
    }

int McuDataState::wormTeeth() const
    {
    return McuAnsHelper(m_aState).word(16);
    }


/********************************************************************/
/*!
    \class McuDataPecState
*/

McuDataPecState::McuDataPecState()
:   m_bPec(false),
    m_bIntellyTrack(false)
    {
    }

bool McuDataPecState::mcuGet(McuSerial &rPort)
    {
    m_aState.clear();
    McuCmdFixed oCmd(
        QString::fromUtf8("#:VS#").toLatin1().constData(), 10,
        QString::fromUtf8("GetPecState"));
    m_oTimeBeg = QTime::currentTime();
    if (!rPort.send(&oCmd))
        return false;
    //m_oTimeEnd = QTime::currentTime();
    m_aState = oCmd.answer();

    McuAnsHelper oAns(m_aState);
    m_bPec = ((oAns.byte(5) & 0x02) != 0);
    m_bIntellyTrack = ((oAns.byte(0) & 0x02) != 0);

    return true;
    }

bool McuDataPecState::mcuSetPec(McuSerial &rPort)
    {
    const char *szCmd = 0;
    switch (m_bPec)
        {
        case true:  szCmd = "#1P"; break;
        case false: szCmd = "#0P"; break;
        };
    if (szCmd == 0)
        return false;
    McuCmdVoid oCmd(
        QString::fromUtf8(szCmd).toLatin1().constData(),
        QString::fromUtf8("SetPEC"));
    if (!rPort.send(&oCmd))
        return false;
    return true;
    }

bool McuDataPecState::mcuSetIntelly(McuSerial &rPort)
    {
    const char *szCmd = 0;
    switch (m_bIntellyTrack)
        {
        case true:  szCmd = "#>I"; break;
        case false: szCmd = "#<I"; break;
        };
    if (szCmd == 0)
        return false;
    // if successful, SetIntellyTrack answers '?' when switched off ans '!' when on
    McuCmdBool oCmd(
        QString::fromUtf8(szCmd).toLatin1().constData(),
        QString::fromUtf8("SetIntellyTrack"), m_bIntellyTrack? '!': '?');
    if (!rPort.send(&oCmd))
        return false;
    if (!oCmd.success())
        LOG(0, "McuDataPecState::mcuSet(): SetIntellyTrack %s returned error \"%s\"",
            m_bIntellyTrack? "on": "off", LQS(oCmd.errorReason()));
    // hack:
    // get sidereal time to avoid troubles with certain following commands
    // (such as <#:VX#)
    McuCmdString oDummy(
        QString::fromUtf8("#:GS#").toLatin1().constData(),
        QString::fromUtf8("GetSiderealTime"));
    rPort.send(&oDummy);
    return true;
    }

QString McuDataPecState::state() const
    {
    McuAnsHelper oAns(m_aState);
    int nState = oAns.byte(5);
    bool bOk = ((nState & 0x80) != 0);
    bool bOnHandbox = ((nState & 0x01) != 0);
    bool bOnSerial = ((nState & 0x02) != 0);
    bool bTraining = ((nState & 0x04) != 0);
    return QString::fromUtf8("%1, %2%3%4 %5 %6 %7 %8 %9 %10 %11 %12")
        .arg(QString::fromUtf8(bOk? "ok": "not ok"))
        .arg(QString::fromUtf8(bTraining? "teach": (bOnHandbox || bOnSerial)? "on": "off"))
        .arg(QString::fromUtf8(bOnHandbox? " hand": ""))
        .arg(QString::fromUtf8(bOnSerial? " serial": ""))
        .arg(oAns.byte(0), 2, 16, QLatin1Char('0'))
        .arg(oAns.word(1), 4, 16, QLatin1Char('0'))
        .arg(oAns.word(3), 4, 16, QLatin1Char('0'))
        .arg(nState, 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(6), 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(7), 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(8), 2, 16, QLatin1Char('0'))
        .arg(oAns.byte(9), 2, 16, QLatin1Char('0'));
    }

int McuDataPecState::pec() const
    {
    McuAnsHelper oAns(m_aState);
    int nState = oAns.byte(5);
    bool bOnHandbox = ((nState & 0x01) != 0);
    bool bOnSerial = ((nState & 0x02) != 0);
    bool bTraining = ((nState & 0x04) != 0);
    return bTraining? 3: bOnHandbox? 2: bOnSerial? 1: 0;
    }

void McuDataPecState::setPec(bool bOn)
    {
    m_bPec = bOn;
    }

bool McuDataPecState::intellyTrack() const
    {
    return m_bIntellyTrack;
    }

void McuDataPecState::setIntellyTrack(bool bOn)
    {
    m_bIntellyTrack = bOn;
    }

double McuDataPecState::wormPos() const
    {
    McuAnsHelper oAns(m_aState);
    int nPecDiv = oAns.word(1);
    int nPecPos = oAns.word(3);
    int nPecPoint = ((oAns.byte(6) & 0x3f) << 2) + (oAns.byte(7) & 0x3);
    return (double)nPecPoint + (double)nPecPos / (double)nPecDiv;
    }


/********************************************************************/
/*!
    \class McuDataPecTable
*/

McuDataPecTable::McuDataPecTable()
    {
    m_nFactor = 1;
    m_aeVal[0] = PecSample::reset;
    for (int nRow = 1; nRow < c_nRowNum; nRow++)
        m_aeVal[nRow] = PecSample::keep;
    }

bool McuDataPecTable::mcuGetFactor(McuSerial &rPort)
    {
    McuCmdFixed oCmd(
        QString::fromUtf8("#:VS#").toLatin1().constData(), 10,
        QString::fromUtf8("GetPecFactor"));
    if (!rPort.send(&oCmd))
        return false;
    m_nFactor = McuAnsHelper(oCmd.answer()).byte(9);
    return true;
    }

bool McuDataPecTable::mcuGetValue(McuSerial &rPort, int nRow)
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;
    McuCmdFixed oCmd(
        QString::fromUtf8("#%1:VR#").arg(nRow).toLatin1().constData(), 1,
        QString::fromUtf8("GetPecValue"));
    if (!rPort.send(&oCmd))
        return false;
    m_aeVal[nRow] = PecSample::int2Val(McuAnsHelper(oCmd.answer()).byte(0) - '0');
    return true;
    }


bool McuDataPecTable::mcuSetFactor(McuSerial &rPort) const
    {
    McuCmdVoid oCmd1(
        QString::fromUtf8("#%1p#").arg(m_nFactor).toLatin1().constData(),
        QString::fromUtf8("SetPecFactor"));
    McuCmdVoid oCmd2(
        QString::fromUtf8("Y").toLatin1().constData(),
        QString::fromUtf8("SaveToEEPROM"));
    if ((!rPort.send(&oCmd1)) || (!rPort.send(&oCmd2)))
        return false;
    return true;
    }

bool McuDataPecTable::mcuSetValue(McuSerial &rPort, int nRow) const
    {
    if ((nRow < 0) || (nRow >= c_nRowNum))
        return false;
    int nVal = PecSample::val2Int(m_aeVal[nRow]);
    McuCmdVoid oCmd(
        QString::fromUtf8("#%1:XR%2#").arg(nRow).arg(nVal).toLatin1().constData(),
        QString::fromUtf8("SetPecValue"));
    if (!rPort.send(&oCmd))
        return false;
    return true;
    }

bool McuDataPecTable::mcuSave(McuSerial &rPort) const
    {
    McuCmdVoid oCmd(
        QString::fromUtf8("#:XYR#").toLatin1().constData(),
        QString::fromUtf8("SavePecTable"));
    if (!rPort.send(&oCmd))
        return false;
    return true;
    }

PecData McuDataPecTable::data(PecData::Traversal eDirection) const
    {
    PecData oData;
    oData.setTraversal(eDirection);
    oData.setFactor(m_nFactor);
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        oData.setValue(nRow, m_aeVal[nRow]);
    oData.setFinish();
    return oData;
    }

bool McuDataPecTable::set(const PecData &oData)
    {
    m_nFactor = oData.factor();
    for (int nRow = 0; nRow < c_nRowNum; nRow++)
        m_aeVal[nRow] = oData[nRow].m_eVal;
    return true;
    }


/********************************************************************/
/*!
    \class McuDataTracking

    \todo SetIntellyTrack to "off" gives back a answer byte '?'.
        Is that harmful?
*/

McuDataTracking::McuDataTracking()
:   m_eRate(McuDataTracking::sidereal)
    {
    }

bool McuDataTracking::mcuSet(McuSerial &rPort)
    {
    const char *szCmd = 0;
    switch (m_eRate)
        {
        case off:      szCmd = "#-1T"; break;
        case sidereal: szCmd = "#0T"; break;
        case solar:    szCmd = "#1T"; break;
        case lunar:    szCmd = "#2T"; break;
        };
    if (szCmd == 0)
        return false;
    McuCmdVoid oCmd(
        QString::fromUtf8(szCmd).toLatin1().constData(),
        QString::fromUtf8("SetTrackingRate"));
    if (!rPort.send(&oCmd))
        return false;
    return true;
    }

McuDataTracking::Rate McuDataTracking::rate() const
    {
    return m_eRate;
    }

void McuDataTracking::setRate(McuDataTracking::Rate eRate)
    {
    m_eRate = eRate;
    }



#if 0

/********************************************************************/
/*!
    \class McuDataIntellyTrack
*/


void McuCom::cmdSetGuideRate(RateSlow eRate)
    {
    const char *szCmd = 0;
    switch (eRate)
        {
        case x1_33: szCmd = ">#0G"; break;
        case x1_66: szCmd = "<#-1G"; break;
        case x2:    szCmd = "<#1G"; break;
        };
    sendCmd(szCmd);
    }


void McuCom::cmdSetMoveRate(RateFast eRate)
    {
    const char *szCmd = 0;
    switch (eRate)
        {
        case x4:  szCmd = "#0:Ss1#"; break;
        case x8:  szCmd = "#1:Ss1#"; break;
        case x16: szCmd = "#2:Ss1#"; break;
        case x20: szCmd = "#3:Ss1#"; break;
        case x27: szCmd = "#4:Ss1#"; break;
        case x40: szCmd = "#5:Ss1#"; break;
        };
    sendCmd(szCmd);
    }


void McuCom::cmdSetSlewRate(RateFast eRate)
    {
    const char *szCmd = 0;
    switch (eRate)
        {
        case x4:  szCmd = "#0:Ss2#"; break;
        case x8:  szCmd = "#1:Ss2#"; break;
        case x16: szCmd = "#2:Ss2#"; break;
        case x20: szCmd = "#3:Ss2#"; break;
        case x27: szCmd = "#4:Ss2#"; break;
        case x40: szCmd = "#5:Ss2#"; break;
        };
    sendCmd(szCmd);
    }


void McuCom::cmdSetSpeed(Speed eSpeed)
    {
    // choose one of the basic speed classes
    // guide (2x): 1.33x, 1.66x, 2x
    // move  (8x): 
    // slew (16x)
    const char *szCmd = 0;
    switch (eSpeed)
        {
        case guide_x2: szCmd = "#:RG#"; break;
        case move_x8:  szCmd = "#:RM#"; break;
        case slew_x16: szCmd = "#:RS#"; break;
        };
    sendCmd(szCmd);
    }


void McuCom::cmdPulseGuide(Direction eDirection, unsigned int nTimeMsecs)
    {
    if (nTimeMsecs > 9999)
        {
        LOG(0, "guiding pulse limited to 9.999 seconds");
        nTimeMsecs = 9999;
        }
    QString qsDirection;
    switch (eDirection)
        {
        case north: qsDirection = "n"; break;
        case south: qsDirection = "s"; break;
        case east:  qsDirection = "e"; break;
        case west:  qsDirection = "w"; break;
        };
    QString qsCmd = QString::fromUtf8("#:Mg%1%2#")
        .arg(qsDirection).arg(nTimeMsecs, 4, 10, QLatin1Char('0'));
    sendCmd(qsCmd.toLatin1().constData());
    }







void McuCom::cmdQueryObjectCoords()
    {
    char acAns[9];
    sendCmd("<#:Gr#", 8, acAns); acAns[8] = '\0';
    LOG(3, "object RA %s", LOG_QS(QString::fromLatin1(acAns)));
    sendCmd("<#:Gd#", 7, acAns); acAns[7] = '\0';
    // translate degree symbol from LX200 (0xdf) to ASCII (0xb0)
    if (acAns[3] == -33)
        acAns[3] = -80;
    LOG(3, "object Dec %s", LOG_QS(QString::fromLatin1(acAns)));
    }


bool McuCom::cmdSetObjectCoords(int nRaHour, int nRaMin, int nRaMinTenth,
    int nDecSign, int nDecDeg, int nDecMin)
    {
    if ((nRaHour < 0) || (nRaHour > 23) ||
        (nRaMin  < 0) || (nRaMin  > 59) ||
        (nRaMinTenth < 0) || (nRaMinTenth > 9) ||
        (nDecDeg < 0) || (nDecDeg > 90) ||
        (nDecMin < 0) || (nDecMin > 59))
        {
        LOG(0, "wrong coords %02d:%02d.%01d %c%02d°%02d",
            nRaHour, nRaMin, nRaMinTenth,
            (nDecSign < 0)? '-': '+', nDecDeg, nDecMin);
        return false;
        }
    char acAns[1];
    QString qsCmd = QString::fromUtf8("#:Sr%1:%2.%3#")
        .arg(nRaHour, 2, 10, QLatin1Char('0'))
        .arg(nRaMin, 2, 10, QLatin1Char('0'))
        .arg(nRaMinTenth, 1, 10, QLatin1Char('0'));
    sendCmd(qsCmd.toLatin1().constData(), 1, acAns);
    if (acAns[0] != '1')
        {
        LOG(0, "failed to send RA %02d:%02d.%01d %c%02d°%02d",
            nRaHour, nRaMin, nRaMinTenth,
            (nDecSign < 0)? '-': '+', nDecDeg, nDecMin);
        return false;
        }
    qsCmd = QString::fromUtf8("#:Sd%1%2%3%4#")
        .arg(QLatin1Char((nDecSign < 0)? '-': '+'))
        .arg(nDecDeg, 2, 10, QLatin1Char('0'))
        .arg(QLatin1Char(-33))
        .arg(nDecMin, 2, 10, QLatin1Char('0'));
    sendCmd(qsCmd.toLatin1().constData(), 1, acAns);
    if (acAns[0] != '1')
        {
        LOG(0, "failed to send Dec %02d:%02d.%01d %c%02d°%02d",
            nRaHour, nRaMin, nRaMinTenth,
            (nDecSign < 0)? '-': '+', nDecDeg, nDecMin);
        return false;
        }

    return true;
    }


void McuCom::cmdQueryTelescopeCoords()
    {
    char acAns[9];
    sendCmd("<#:GR#", 8, acAns); acAns[8] = '\0';
    LOG(3, "telescope RA %s", QString::fromLatin1(acAns).toUtf8().constData());
    sendCmd("<#:GD#", 7, acAns); acAns[7] = '\0';
    // translate degree symbol from LX200 (0xdf) to ASCII (0xb0)
    if (acAns[3] == -33)
        acAns[3] = -80;
    LOG(3, "telescope Dec %s", QString::fromLatin1(acAns).toUtf8().constData());
    }


void McuCom::cmdSyncTelescopeToObject()
    {
    char acAns[1];
    sendCmd("#:MS#", 1, acAns);
    }


bool McuCom::cmdSlewTelescopeToObject()
    {
    char acAns[1];
    sendCmd("#:MS#", 1, acAns);
    if (acAns[0] != '0')
        {
        // todo: treat further bytes, e.g. " No Goto#"
        LOG(0, "failed to slew to object");
        return false;
        }
    return true;
    }


#endif
